#
# Revelation 0.1.0 - a password manager for GNOME 2
# http://oss.wired-networks.net/revelation/
#
# Module containing data-related functionality
#
#
# Copyright (c) 2003-2004 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#


import gtk, revelation, gobject, libxml2, string
from Crypto.Cipher import AES

DATASTORE_COL_NAME = 0
DATASTORE_COL_ICON = 1
DATASTORE_COL_TYPE = 2
DATASTORE_COL_DESC = 3
DATASTORE_COL_UPDATED = 4
DATASTORE_COL_FIELDS = 5


ENTRY_FOLDER			= "folder"
ENTRY_ACCOUNT_CREDITCARD	= "creditcard"
ENTRY_ACCOUNT_DATABASE		= "database"
ENTRY_ACCOUNT_DOOR		= "door"
ENTRY_ACCOUNT_EMAIL		= "email"
ENTRY_ACCOUNT_FTP		= "ftp"
ENTRY_ACCOUNT_GENERIC		= "generic"
ENTRY_ACCOUNT_PHONE		= "phone"
ENTRY_ACCOUNT_SHELL		= "shell"
ENTRY_ACCOUNT_USENET		= "usenet"
ENTRY_ACCOUNT_WEBSITE		= "website"


FIELD_GENERIC_CODE		= "generic-code"
FIELD_GENERIC_EMAIL		= "generic-email"
FIELD_GENERIC_HOSTNAME		= "generic-hostname"
FIELD_GENERIC_LOCATION		= "generic-location"
FIELD_GENERIC_PASSWORD		= "generic-password"
FIELD_GENERIC_PIN		= "generic-pin"
FIELD_GENERIC_PORT		= "generic-port"
FIELD_GENERIC_URL		= "generic-url"
FIELD_GENERIC_USERNAME		= "generic-username"

FIELD_CREDITCARD_CARDTYPE	= "creditcard-cardtype"
FIELD_CREDITCARD_CARDNUMBER	= "creditcard-cardnumber"
FIELD_CREDITCARD_CCV		= "creditcard-ccv"
FIELD_CREDITCARD_EXPIRYDATE	= "creditcard-expirydate"

FIELD_PHONE_PHONENUMBER		= "phone-phonenumber"


FIELD_TYPE_EMAIL		= "email"
FIELD_TYPE_PASSWORD		= "password"
FIELD_TYPE_TEXT			= "text"
FIELD_TYPE_URL			= "url"

ENCRYPT_KEYSIZE			= 32


class EntryTypes(object):

	def __init__(self):

		self.entrytype = {

			ENTRY_FOLDER			: {
				"name"		: "Folder",
				"icon"		: revelation.ui.STOCK_FOLDER,
				"fields"	: []
			},

			ENTRY_ACCOUNT_CREDITCARD	: {
				"name"		: "Creditcard",
				"icon"		: revelation.ui.STOCK_ACCOUNT_CREDITCARD,
				"fields"	: [ FIELD_CREDITCARD_CARDTYPE, FIELD_CREDITCARD_CARDNUMBER, FIELD_CREDITCARD_EXPIRYDATE, FIELD_CREDITCARD_CCV, FIELD_GENERIC_PIN ]
			},

			ENTRY_ACCOUNT_DATABASE		: {
				"name"		: "Database",
				"icon"		: revelation.ui.STOCK_ACCOUNT_DATABASE,
				"fields"	: [ FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_DOOR		: {
				"name"		: "Door lock",
				"icon"		: revelation.ui.STOCK_ACCOUNT_DOOR,
				"fields"	: [ FIELD_GENERIC_LOCATION, FIELD_GENERIC_CODE ]
			},

			ENTRY_ACCOUNT_EMAIL		: {
				"name"		: "Email",
				"icon"		: revelation.ui.STOCK_ACCOUNT_EMAIL,
				"fields"	: [ FIELD_GENERIC_EMAIL, FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_FTP		: {
				"name"		: "FTP",
				"icon"		: revelation.ui.STOCK_ACCOUNT_FTP,
				"fields"	: [ FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_PORT, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_GENERIC		: {
				"name"		: "Generic",
				"icon"		: revelation.ui.STOCK_ACCOUNT,
				"fields"	: [ FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_PHONE		: {
				"name"		: "Phone",
				"icon"		: revelation.ui.STOCK_ACCOUNT_PHONE,
				"fields"	: [ FIELD_PHONE_PHONENUMBER, FIELD_GENERIC_PIN ]
			},

			ENTRY_ACCOUNT_SHELL		: {
				"name"		: "Shell",
				"icon"		: revelation.ui.STOCK_ACCOUNT_SHELL,
				"fields"	: [ FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_USENET		: {
				"name"		: "Usenet",
				"icon"		: revelation.ui.STOCK_ACCOUNT_USENET,
				"fields"	: [ FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_WEBSITE		: {
				"name"		: "Website",
				"icon"		: revelation.ui.STOCK_ACCOUNT_WEBSITE,
				"fields"	: [ FIELD_GENERIC_URL, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			}
		}

		self.fieldtype = {

			FIELD_GENERIC_CODE		: { "name" : "Code",		"type" : FIELD_TYPE_PASSWORD },
			FIELD_GENERIC_EMAIL		: { "name" : "Email address",	"type" : FIELD_TYPE_EMAIL },
			FIELD_GENERIC_HOSTNAME		: { "name" : "Hostname",	"type" : FIELD_TYPE_TEXT },
			FIELD_GENERIC_LOCATION		: { "name" : "Location",	"type" : FIELD_TYPE_TEXT },
			FIELD_GENERIC_PASSWORD		: { "name" : "Password",	"type" : FIELD_TYPE_PASSWORD },
			FIELD_GENERIC_PIN		: { "name" : "PIN",		"type" : FIELD_TYPE_PASSWORD },
			FIELD_GENERIC_PORT		: { "name" : "Port number",	"type" : FIELD_TYPE_TEXT },
			FIELD_GENERIC_URL		: { "name" : "URL",		"type" : FIELD_TYPE_URL },
			FIELD_GENERIC_USERNAME		: { "name" : "Username",	"type" : FIELD_TYPE_TEXT },

			FIELD_CREDITCARD_CARDTYPE	: { "name" : "Card type",	"type" : FIELD_TYPE_TEXT },
			FIELD_CREDITCARD_CARDNUMBER 	: { "name" : "Card number",	"type" : FIELD_TYPE_TEXT },
			FIELD_CREDITCARD_CCV		: { "name" : "CCV number",	"type" : FIELD_TYPE_TEXT },
			FIELD_CREDITCARD_EXPIRYDATE	: { "name" : "Expiry date",	"type" : FIELD_TYPE_TEXT },

			FIELD_PHONE_PHONENUMBER		: { "name" : "Phone number",	"type" : FIELD_TYPE_TEXT }
		}


	def get_field_data(self, field):
		"""Fetches data for a field type"""

		return self.fieldtype[field]


	def get_list(self):
		"""Fetches a list of entry types"""

		list = self.entrytype.keys()
		list.sort()

		return list


	def get_data(self, type):
		"""Returns data for a type"""

		return self.entrytype[type]


class DataStore(gtk.TreeStore):

	def __init__(self, entrytypes):
		"""Sets up the datastore"""

		gtk.TreeStore.__init__(self, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_INT, gobject.TYPE_PYOBJECT)
		self.entrytypes = entrytypes
		self.changed = gtk.FALSE


	def add_entry(self, parent, data = None):
		"""Adds an entry to the data store"""

		iter = self.append(parent)
		self.changed = gtk.TRUE

		if data != None:
			self.update_entry(iter, data)

		return iter


	def clear(self):
		"""Removes all data from the datastore"""

		gtk.TreeStore.clear(self)
		self.changed = gtk.FALSE


	def export_xml(self, parent = None, level = 0):
		"""Generates xml based on data in the datastore"""

		xml = ""

		# process each child
		numchildren = self.iter_n_children(parent)
		for i in range(numchildren):

			# get child data
			iter = self.iter_nth_child(parent, i)
			data = self.get_entry(iter)

			tabs = "\t" * (level + 1)

			# generate entry
			xml = xml + "\n"
			xml = xml + tabs + "<entry type=\"" + data["type"] + "\">\n"
			xml = xml + tabs + "	<name>" + data["name"] + "</name>\n"
			xml = xml + tabs + "	<description>" + data["description"] + "</description>\n"
			xml = xml + tabs + "	<updated>" + str(data["updated"]) + "</updated>\n"

			# generate fields
			for field in self.entrytypes.get_data(data["type"])["fields"]:
				xml = xml + tabs + "	<field id=\"" + field + "\">" + data["fields"][field] + "</field>\n"

			# handle any children
			if self.iter_n_children(iter) > 0:
				xml = xml + self.export_xml(iter, level + 1)

			xml = xml + tabs + "</entry>\n"

		# generate header and footer if at level 0
		if level == 0:
			header = "<?xml version=\"1.0\" ?>\n"
			header = header + "<revelationdata version=\"" + revelation.VERSION + "\">\n"
			footer = "</revelationdata>\n"

			xml = header + xml + footer

		return xml


	def get_entry(self, iter):
		"""Fetches data for a tree entry"""

		data = {
			"name"		: self.get_value(iter, DATASTORE_COL_NAME),
			"type"		: self.get_value(iter, DATASTORE_COL_TYPE),
			"icon"		: self.get_value(iter, DATASTORE_COL_ICON),
			"description"	: self.get_value(iter, DATASTORE_COL_DESC),
			"updated"	: self.get_value(iter, DATASTORE_COL_UPDATED),
			"fields"	: self.get_value(iter, DATASTORE_COL_FIELDS)
		}

		# always return the normal, closed folder icon for folders.
		# the open folder icon is only for tree-internal use.
		if data["icon"] == revelation.ui.STOCK_FOLDER_OPEN:
			data["icon"] = revelation.ui.STOCK_FOLDER

		return data


	def get_entry_type(self, iter):
		"""Returns the type of an entry"""

		if iter == None:
			return None
		else:
			return self.get_value(iter, DATASTORE_COL_TYPE)


	def import_xml(self, xml):
		"""Imports data from xml"""

		try: doc = libxml2.parseDoc(xml)
		except TypeError:
			return gtk.FALSE

		# fetch and validate root
		root = doc.children
		if root.name != "revelationdata":
			return gtk.FALSE

		# process entries
		child = root.children
		while child is not None:
			if self.import_xml_node(child) == gtk.FALSE:
				return gtk.FALSE
			child = child.next

		return gtk.TRUE


	def import_xml_node(self, node, parent = None):
		"""Adds entries to the datastore recursively"""

		# skip text outside nodes
		if node.type == "text":
			return gtk.TRUE

		# skip unknown nodes
		if node.type != "element" or node.name != "entry":
			return gtk.FALSE

		# set up empty data structure
		data = {
			"name"		: "",
			"type"		: self.import_xml_get_attrs(node)["type"],
			"description"	: "",
			"updated"	: 0,
			"fields"	: {}
		}

		# add empty entry, iter needed for any children
		iter = self.add_entry(parent)

		# get elements
		child = node.children
		while child is not None:

			if child.type == "element":

				if child.name == "name":
					data["name"] = child.content

				elif child.name == "description":
					data["description"] = child.content

				elif child.name == "updated":
					data["updated"] = int(child.content)

				elif child.name == "field":
					field = self.import_xml_get_attrs(child)["id"]
					data["fields"][field] = child.content

				elif child.name == "entry":
					self.import_xml_node(child, iter)

				else:
					return gtk.FALSE

			child = child.next

		# set up missing fields according to entry type
		for field in self.entrytypes.get_data(data["type"])["fields"]:
			if data["fields"].has_key(field) == False:
				data["fields"][field] = ""

		# update entry with actual data
		self.update_entry(iter, data)

		return gtk.TRUE


	def import_xml_get_attrs(self, node):
		"""Returns a dict of attributes for a node"""

		attrs = {}

		attr = node.properties
		while attr is not None:
			attrs[attr.name] = attr.content
			attr = attr.next

		return attrs


	def move_entry(self, iter, dest):
		"""Moves an entry recursively to a new parent"""

		# create the entry at the new parent
		newiter = self.add_entry(dest, self.get_entry(iter))

		# handle any children
		while self.iter_n_children(iter) > 0:
			self.move_entry(self.iter_nth_child(iter, 0), newiter)

		parent = self.iter_parent(iter)

		# remove the entry
		self.remove_entry(iter)

		# collapse parent if empty
		if self.iter_n_children(parent) == 0:
			self.set_folder_state(parent, gtk.FALSE)

		return newiter


	def remove_entry(self, iter):
		"""Removes an entry, recursive if folder"""

		# remove children
		while self.iter_n_children(iter) > 0:
			self.remove_entry(self.iter_children(iter))

		parent = self.iter_parent(iter)

		# remove the entry
		self.remove(iter)

		# collapse parent if empty
		if self.iter_n_children(parent) == 0:
			self.set_folder_state(parent, gtk.FALSE)

		self.changed = gtk.TRUE


	def set_folder_state(self, iter, open):
		"""Sets the open/closed state of a folder, changing its icon"""

		if iter == None:
			return

		if self.get_entry_type(iter) != ENTRY_FOLDER:
			return

		if open == gtk.TRUE:
			icon = revelation.ui.STOCK_FOLDER_OPEN
		else:
			icon = revelation.ui.STOCK_FOLDER

		self.set_value(iter, DATASTORE_COL_ICON, icon)


	def update_entry(self, iter, data):
		"""Updates data for an entry"""

		self.set_value(iter, DATASTORE_COL_NAME, data["name"])
		self.set_value(iter, DATASTORE_COL_TYPE, data["type"])
		self.set_value(iter, DATASTORE_COL_DESC, data["description"])
		self.set_value(iter, DATASTORE_COL_UPDATED, data["updated"])
		self.set_value(iter, DATASTORE_COL_FIELDS, data["fields"])

		# keep icon if current is folder-open and the type still is folder
		if self.get_value(iter, DATASTORE_COL_ICON) != revelation.ui.STOCK_FOLDER_OPEN or data["type"] != ENTRY_FOLDER:
			self.set_value(iter, DATASTORE_COL_ICON, self.entrytypes.get_data(data["type"])["icon"])

		self.changed = gtk.TRUE


class FolderStore(gtk.TreeStore):
	"""A folder store based on a datastore"""

	def __init__(self, datastore):

		gtk.TreeStore.__init__(self, gobject.TYPE_STRING, gobject.TYPE_STRING)
		self.datastore = datastore

		self.import_folders(self.datastore)


	def add_folder(self, parent, name):
		"""Adds a folder to the folderstore"""

		iter = self.append(parent)
		self.set_value(iter, DATASTORE_COL_NAME, name)
		self.set_value(iter, DATASTORE_COL_ICON, revelation.ui.STOCK_FOLDER)

		return iter


	def get_iter_from_orig(self, origiter):
		"""Returns the folder iter corresponding to the passed iter in the original datastore"""

		if origiter == None:
			return None

		datapath = self.datastore.get_path(origiter)
		dataiter = None
		folderiter = None

		# go through the path
		for pathitem in datapath:

			# find the current folders index in the datastore
			folderindex = -1

			for n in range(pathitem + 1):
				iter = self.datastore.iter_nth_child(dataiter, n)

				if self.datastore.get_entry_type(iter) == ENTRY_FOLDER:
					folderindex = folderindex + 1

			# find the corresponding folder in the folderstore
			folderiter = self.iter_nth_child(folderiter, folderindex)

		return folderiter


	def get_orig_iter(self, iter):
		"""Returns an iter from the original datastore corresponding to this iter from the folderstore"""

		if iter == None:
			return None

		folderpath = self.get_path(iter)
		dataiter = None

		# go through the path, looking for the iter
		for pathitem in folderpath:

			folderindex = -1

			# scan all nodes at this level
			for dataindex in range(self.datastore.iter_n_children(dataiter)):

				child = self.datastore.iter_nth_child(dataiter, dataindex)

				if self.datastore.get_entry_type(child) == ENTRY_FOLDER:
					folderindex = folderindex + 1

				if folderindex == pathitem:
					dataiter = child
					break

		return dataiter


	def import_folders(self, datastore, dataparent = None, folderparent = None):
		"""Imports folders from a datastore"""

		for i in range(datastore.iter_n_children(dataparent)):
			child = datastore.iter_nth_child(dataparent, i)

			data = datastore.get_entry(child)
			if data["type"] == ENTRY_FOLDER:
				folder = self.add_folder(folderparent, data["name"])
				self.import_folders(datastore, child, folder)


	def set_folder_state(self, iter, open):
		"""Sets the open/closed state of a folder, changing its icon"""

		if iter == None:
			return

		if open == gtk.TRUE:
			icon = revelation.ui.STOCK_FOLDER_OPEN
		else:
			icon = revelation.ui.STOCK_FOLDER

		self.set_value(iter, DATASTORE_COL_ICON, icon)


def decrypt(data, password):
	"""decrypts data"""

	padlen = ENCRYPT_KEYSIZE - (len(password) % ENCRYPT_KEYSIZE)
	AES.key_size = ENCRYPT_KEYSIZE
	cipher = AES.new(password + (chr(0) * padlen))

	if len(data) % cipher.block_size != 0:
		return ""

	decrypted = cipher.decrypt(data)

	return decrypted


def encrypt(data, password):
	"""encrypts data"""

	padlen = ENCRYPT_KEYSIZE - (len(password) % ENCRYPT_KEYSIZE)
	AES.key_size = ENCRYPT_KEYSIZE
	cipher = AES.new(password + (chr(0) * padlen))

	padlen = cipher.block_size - (len(data) % cipher.block_size)
	if padlen == cipher.block_size:
		padlen = 0
	encrypted = cipher.encrypt(data + (" " * padlen))

	return encrypted


def file_read(file):
	"""reads data from a file"""

	try: fp = open(file, "rb", 0)
	except IOError:
		return None

	data = fp.read()
	fp.close()

	return data


def file_write(file, data):
	"""writes data to a file"""

	try: fp = open(file, "wb", 0)
	except IOError:
		return gtk.FALSE

	fp.write(data)
	fp.flush()
	fp.close()

	return gtk.TRUE

