#
# Revelation 0.1.0 - a password manager for GNOME 2
# http://oss.wired-networks.net/revelation/
#
# Module containing ui-related functionality
#
#
# Copyright (c) 2003-2004 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

import gtk, gtk.gdk, gnome.ui, revelation, time, os

STOCK_ACCOUNT_CREDITCARD = "revelation-account-creditcard"
STOCK_ACCOUNT_DATABASE = "revelation-account-database"
STOCK_ACCOUNT_DOOR = "revelation-account-door"
STOCK_ACCOUNT_EMAIL = "revelation-account-email"
STOCK_ACCOUNT_FTP = "revelation-account-ftp"
STOCK_ACCOUNT_GENERIC = "revelation-account-generic"
STOCK_ACCOUNT_PHONE = "revelation-account-phone"
STOCK_ACCOUNT_SHELL = "revelation-account-shell"
STOCK_ACCOUNT_USENET = "revelation-account-usenet"
STOCK_ACCOUNT_WEBSITE = "revelation-account-website"

STOCK_ACCOUNT = STOCK_ACCOUNT_GENERIC
STOCK_APPLICATION = "revelation-application"
STOCK_FOLDER = "revelation-folder"
STOCK_FOLDER_OPEN = "revelation-folder-open"
STOCK_PASSWORD = "revelation-password"

ICON_SIZE_DATAVIEW = gtk.icon_size_register("revelation-dataview", 24, 24)
ICON_SIZE_DROPDOWN = gtk.icon_size_register("revelation-dropdown", 16, 16)

# this one needs to be a normal gtk icon size, as the tree cellrenderer
# seems to have problems with custom ones.
ICON_SIZE_TREEVIEW = gtk.ICON_SIZE_SMALL_TOOLBAR


class DataView(gtk.VBox):

	def __init__(self, entrytypes):

		gtk.VBox.__init__(self)
		self.set_spacing(5)
		self.set_border_width(10)

		self.infoview = gtk.FALSE
		self.entrytypes = entrytypes


	def ui_field_label(self, field):
		"""Renders a field label"""

		label = gtk.Label()
		label.set_markup("<b>" + self.entrytypes.get_field_data(field)["name"] + ":</b>")
		label.set_alignment(1, 0.5)

		return label


	def ui_field_value(self, field, value):
		"""Renders a field, returns the widget created"""

		type = self.entrytypes.get_field_data(field)["type"]

		if type == revelation.data.FIELD_TYPE_EMAIL:
			widget = gnome.ui.HRef("mailto:" + value, value)

		elif type == revelation.data.FIELD_TYPE_URL:
			widget = gnome.ui.HRef(value, value)

		else:
			widget = gtk.Label(value)
			widget.set_alignment(0, 0.5)
			widget.set_selectable(gtk.TRUE)
			
		return widget


	def clear(self, force = gtk.FALSE):
		"""Clears all contents of the dataview"""

		if self.infoview == gtk.FALSE or force == gtk.TRUE:
			for child in self.get_children():
				child.destroy()


	def display_entry(self, data):
		"""Displays an account entry"""

		self.clear(gtk.TRUE)
		self.infoview = gtk.FALSE

		# set up icon and name
		hbox = gtk.HBox()
		hbox.set_spacing(5)
		self.pack_start(hbox)

		image = gtk.Image()
		image.set_from_stock(data["icon"], ICON_SIZE_DATAVIEW)
		hbox.pack_start(image)

		title = gtk.Label()
		title.set_markup("<span size=\"large\" weight=\"bold\">" + data["name"] + "</span>")
		hbox.pack_start(title)

		# description
		desc = gtk.Label()
		desc.set_use_markup(gtk.TRUE)
		desc.set_line_wrap(gtk.TRUE)

		desctext = "<span weight=\"bold\">" + self.entrytypes.get_data(data["type"])["name"] + "</span>"

		if data["description"] != "":
			desctext = desctext + "<span weight=\"bold\">;</span> " + data["description"]

		desc.set_markup(desctext)
		self.pack_start(desc)

		# render fields, if any
		fieldlist = self.entrytypes.get_data(data["type"])["fields"]
		if len(fieldlist) > 0:

			# set up field table
			table = gtk.Table(len(fieldlist), 2, gtk.FALSE)
			table.set_border_width(10)
			table.set_col_spacings(5)
			table.set_row_spacings(2)
			self.pack_start(table)

			# render fields
			for field, index in zip(fieldlist, range(len(fieldlist))):
				table.attach(self.ui_field_label(field), 0, 1, index, index + 1, gtk.FILL)
				table.attach(self.ui_field_value(field, data["fields"][field]), 1, 2, index, index + 1, gtk.FILL)

			# update-time
			self.pack_start(gtk.Label("Updated " + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(data["updated"]))))

		# display data
		self.show_all()


	def display_info(self):
		"""Displays info about the application"""

		self.clear()
		self.infoview = gtk.TRUE

		# set up "logo" (appname etc)
		hbox = gtk.HBox(gtk.FALSE, 10)
		self.pack_start(hbox)

		image = gtk.Image()
		image.set_from_stock(STOCK_APPLICATION, gtk.ICON_SIZE_DND)
		hbox.pack_start(image, gtk.FALSE, gtk.FALSE)

		label = gtk.Label()
		label.set_markup("<span size=\"x-large\" weight=\"bold\">" + revelation.APPNAME + " " + revelation.VERSION + "</span>")
		hbox.pack_start(label, gtk.FALSE, gtk.FALSE)

		# intro text and href
		self.pack_start(gtk.Label("A password manager for GNOME 2"))

		# gpl
		gpl = "\nThis program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.\n\n"
		gpl = gpl + "This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details."
		label = gtk.Label()
		label.set_line_wrap(gtk.TRUE)
		label.set_size_request(250, -1)
		label.set_markup("<span size=\"x-small\">" + gpl + "</span>")
		self.pack_start(label)


	def pack_start(self, widget):
		"""Adds a widget to the data view"""

		alignment = gtk.Alignment(0.5, 0.5, 0, 0)
		alignment.add(widget)
		gtk.VBox.pack_start(self, alignment)


class IconFactory(gtk.IconFactory):

	def __init__(self):

		gtk.IconFactory.__init__(self)
		self.add_default()

		# names of icon images
		icons = {
			STOCK_APPLICATION		: "revelation.png",
			STOCK_ACCOUNT_CREDITCARD	: "account-creditcard.png",
			STOCK_ACCOUNT_DATABASE		: "account-database.png",
			STOCK_ACCOUNT_DOOR		: "account-door.png",
			STOCK_ACCOUNT_EMAIL		: "account-email.png",
			STOCK_ACCOUNT_FTP		: "account-ftp.png",
			STOCK_ACCOUNT_GENERIC		: "account-generic.png",
			STOCK_ACCOUNT_PHONE		: "account-phone.png",
			STOCK_ACCOUNT_SHELL		: "account-shell.png",
			STOCK_ACCOUNT_USENET		: "account-usenet.png",
			STOCK_ACCOUNT_WEBSITE		: "account-website.png",
			STOCK_FOLDER			: "folder.png",
			STOCK_FOLDER_OPEN		: "folder-open.png",
			STOCK_PASSWORD			: "password.png"
		}

		for id, filename in icons.items():
			iconset = gtk.IconSet(gtk.gdk.pixbuf_new_from_file(revelation.DATADIR + "/pixmaps/" + filename))
			self.add(id, iconset)


class InputSection(gtk.VBox):
	"""An input section, containing rows of input fields"""

	def __init__(self, title = None, sizegroup = None):

		gtk.VBox.__init__(self)
		self.set_border_width(0)
		self.set_spacing(6)

		self.title = title
		
		if title != None:
			label = gtk.Label("<span size=\"larger\" weight=\"bold\">" + title + "</span>")
			label.set_use_markup(gtk.TRUE)
			label.set_alignment(0, 0.5)
			self.pack_start(label, gtk.FALSE)

		if sizegroup == None:
			self.sizegroup = gtk.SizeGroup(gtk.SIZE_GROUP_HORIZONTAL)
		else:
			self.sizegroup = sizegroup


	def add_inputrow(self, title, widget):
		"""Adds an input row to the section"""

		# set up indentation
		if self.title == None:
			indent = ""
		else:
			indent = "    "


		row = gtk.HBox()
		row.set_spacing(6)

		label = gtk.Label(indent + title + ":")
		label.set_alignment(0, 0.5)
		self.sizegroup.add_widget(label)
		row.pack_start(label, gtk.FALSE, gtk.FALSE)

		row.pack_start(widget)

		self.pack_start(row)

		return row


	def clear(self):
		"""Clears the input section, except the label if present"""

		# do not remove the title
		if self.title == None:
			leave = 0
		else:
			leave = 1

		rows = self.get_children()
		while len(rows) > leave:
			rows.pop(leave).destroy()


class Tree(gtk.TreeView):
	"""Creates a tree used as interface to the datastore"""

	def __init__(self, datastore, title):

		gtk.TreeView.__init__(self, datastore)
		self.set_headers_visible(gtk.TRUE)
		self.append_column(self.ui_column(title))

		self.connect("button_press_event", self.cb_buttonpress)
		self.connect("key_press_event", self.cb_keypress)
		self.connect("row-expanded", self.cb_row_expanded)
		self.connect("row-collapsed", self.cb_row_collapsed)


	def ui_column(self, title):
		"""Sets up a column"""

		column = gtk.TreeViewColumn()
		column.set_title(title)

		cr = gtk.CellRendererPixbuf()
		column.pack_start(cr, gtk.FALSE)
		column.add_attribute(cr, "stock-id", revelation.data.DATASTORE_COL_ICON)
		cr.set_property("stock-size", ICON_SIZE_TREEVIEW)

		cr = gtk.CellRendererText()
		column.pack_start(cr, gtk.TRUE)
		column.add_attribute(cr, "text", revelation.data.DATASTORE_COL_NAME)

		return column


	def cb_buttonpress(self, object, data):
		"""Callback for mouseclicks"""

		# expand on lmb doubleclick
		if data.button == 1 and data.type == gtk.gdk._2BUTTON_PRESS:
			path = self.get_path_at_pos(data.x, data.y)
			if path == None:
				return

			self.toggle_expanded(self.get_model().get_iter(path[0]))


	def cb_keypress(self, object, data):
		"""Callback for keypresses"""

		# expand if space is pressed
		if data.keyval == 32:
			self.toggle_expanded(self.get_selected())


	def cb_row_collapsed(self, object, iter, extra):
		self.collapse_row(self.get_model().get_path(iter))


	def cb_row_expanded(self, object, iter, extra):
		self.get_model().set_folder_state(iter, gtk.TRUE)


	def collapse_row(self, path):
		"""Collapses a path, and all its children"""

		model = self.get_model()
		iter = model.get_iter(path)
		numchildren = model.iter_n_children(iter)

		model.set_folder_state(iter, gtk.FALSE)

		if numchildren > 0:
			for i in range(numchildren):
				self.collapse_row(model.get_path(model.iter_nth_child(iter, i)))

		path = self.get_model().get_path(iter)
		gtk.TreeView.collapse_row(self, path)


	def expand_to_iter(self, iter):
		"""Expands all parents of an iter"""

		if iter == None:
			return

		path = self.get_model().get_path(iter)
		for i in range(len(path[:-1])):
			self.expand_row(path[0:i+1], gtk.FALSE)


	def get_selected(self):
		"""Returns the selected row"""
		
		return self.get_selection().get_selected()[1]

	def select(self, iter):
		"""Selects an iter"""

		path = self.get_model().get_path(iter)

		# expand to iter, and select it
		self.expand_to_iter(iter)
		self.set_cursor(path)


	def toggle_expanded(self, iter):
		"""Expands or collapses a row, based on its current status"""

		if iter == None:
			return

		path = self.get_model().get_path(iter)

		if self.row_expanded(path) == gtk.TRUE:
			self.collapse_row(path)
		else:
			self.expand_row(path, gtk.FALSE)

	def unselect_all(self):
		"""Unselects all entries"""

		self.get_selection().unselect_all()
		self.emit("cursor_changed")
		self.emit("unselect_all")

